/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <defs.h>
#include <debug.h>
#include <packet.h>
#include <eventhandler.h>
#include <scheduler.h>
#include <queue.h>
#include <state.h>
#include <codeset.h>
#include <intransit.h>

//Packet class functions
//Parameters of a packet. Also contains functions that allow calculation of the packets position (boundary) at any point in time/space
//and calculation of the state of the line at a point in time/space due to this packet.

extern Codeset *codeset;
extern int propSpeed;
extern Scheduler * sched;
extern InTransit *it;
extern int * nstate;


Packet::Packet(): next(NULL), prev(NULL), id(0)
{

	backoffs =0;
}

Packet::Packet(Packet & p)
{
	
	next = NULL;
	prev = NULL;
	id = p.id;

	srcaddr = p.srcaddr;
	srcport = p.srcport;
	dstaddr = p.dstaddr;
	dstport = p.dstport;
	size = p.size;

	codeword = p.codeword;
	
	depx = p.depx;
	depy = p.depy;
	depz = p.depz;

	departuretime = p.departuretime;;
	arrivaltime = p.arrivaltime;;

	chiprate = p.chiprate;
 	bitrate = p.bitrate;
	
	status = p.status;
	A = p.A;
	N = p.N;
	w = p.w;

	backoffs=p.backoffs;
}

Packet::~Packet()
{
}


void Packet::calcboundary(double *head, double *tail)
{


//Given a packet (i.e. size and departure time), this function calculates the boundary (head/tail) of the packet at the 
//current time. 

//Topology is a flat bus structure. y = z = 0. x measures the distances from the headend.
		dprintf("packet.cc: Packet::calcboundary(): Checking state for packet: %d\n", id); 		
		dprintf("packet.cc: Packet::calcboundary(): Departure time: %.12f\n", departuretime); 		
		dprintf("packet.cc: Packet::calcboundary(): Size: %ld\n", size); 		
		dprintf("packet.cc: Packet::calcboundary(): Bit rate: %.12f\n", bitrate); 		
		
		double transittime = sched->currenttime - departuretime;
		dprintf("packet.cc: Packet::calcboundary(): Transit time: %.12f s\n", transittime); 		
	 	
		//Measured in meters
		double transitdistance = propSpeed * transittime; 
		dprintf("packet.cc: Packet::calcboundary(): Transit distance: %.12f m \n", transitdistance); 		
	

		//Chip length in meters
		double chiplength = (1.0/chiprate) * propSpeed;	
		dprintf("packet.cc: Packet::calcboundary(): Length of chip: %.12f m \n", chiplength); 		
		
		//Neglect any fraction of the transmission time - assume everything is on pico second boundaries
		double packetlength = (double)size * 8.0 * codeset->N;
		dprintf("packet.cc: Packet::calcboundary(): Packet length: %.12f chips\n", packetlength); 		
		
		double transmissionspan = packetlength * chiplength;
		dprintf("packet.cc: Packet::calcboundary(): Transmission span: %.12f m\n", transmissionspan); 		

		*head = depx - transitdistance;
		*tail = *head +  transmissionspan;
		dprintf("packet.cc: Packet::calcboundary(): Packet head: %.12f m\n", *head); 		
		dprintf("packet.cc: Packet::calcboundary(): Packet tail: %.12f m\n", *tail); 		
}






int * Packet::calcstate(double x, double y, double z, double time, int topology )
{

//This function calculates the state of the line due to this packet at 
//(x, y, z) at the current time.
//Note that that state is measured between x and x+dx

	int * state = NULL;
//Topology is a flat bus structure. y = z = 0. x measures the distances from the headend.
	dprintf("packet.cc: Packet::calcstate(): Checking state for packet: %d\n",id); 		

	double packethead;
	double packettail;
	calcboundary(&packethead, &packettail);
	dprintf("packet.cc: Packet::calcstate(): Packet head: %.12f m\n", packethead); 		
	dprintf("packet.cc: Packet::calcstate(): Packet tail: %.12f m\n", packettail); 		

	//Chip length in meters
	double chiplength = (1.0/chiprate) * propSpeed;	
	dprintf("packet.cc: Packet::calcstate(): Length of chip: %.12f m \n", chiplength); 		
	double x1 = x;
	double x2 = x + N*chiplength;
//State is measured from the point x to a point x-dx, where dx corresponds to the length of the codeset (N)
//Each element of the state array represents a chip on the line.

	if (((packethead <= x1 ) && (packettail >= x2)) ||((packethead > x1 ) && (packethead < x2)) ||((packethead >x1 ) && (packethead < x2) && (packettail >= x2)) || ((packethead <= x1 ) && (packettail < x2) && (packettail > x1)))
	{

		dprintf("packet.cc: Packet::calcstate(): Found packet in region:  %d\n", id); 	
		double offset = x1 - packethead;
		dprintf("packet.cc: Packet::calcstate(): Offset: %.6f\n", offset); 	

		int chipoffset =0;
		if (offset >= 0)
			chipoffset = int((offset/chiplength) + 0.5);
		else
			chipoffset = int((offset/chiplength) - 0.5);

		chipoffset = chipoffset % (int) N;	
		if(chipoffset < 0)
			chipoffset = -chipoffset;
		
		dprintf("packet.cc:Packet::calcstate(): Code: \n");
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
		{
				dprintf("%d ", codeset->set[codeword*(A*N) + N*count1  + count]);
		}
		::dprintf("\n");	
		
		dprintf("packet.cc: Packet::calcstate(): Offset in front of head: %d (chips)\n", chipoffset); 		

		if(packethead <= x1)
			state = codeset->LeftShiftCodeword(codeword, chipoffset);
		else
			state = codeset->LeftShiftCodeword(codeword, N - chipoffset);
		
		dprintf("packet.cc: Packet::calcstate(): Packet shifted\n"); 		

		dprintf("packet.cc:Packet::calcstate(): Shifted Code: \n");
		for (int count1 =0; count1 < A; count1++)
			for (int count =N-1; count >=0; count--)
			{
				::dprintf("%d ", state[count1*N + count]);	
			}
		::dprintf("\n");	
	
	}
	else
	{
		dprintf("packet.cc: Packet::calcstate(): Packet not in region:  %d\n", id); 		
	}
	return state;	

}

void Packet::calcnormalizedstate()
{
//Calculates the normalized state of the line due to this packet.
//( wrt a N chip long clock cycle)


	double chiptime = (1.0/chiprate);
	int chips = int((departuretime/chiptime) + 0.5);	
	int chipoffset = (int)chips % (int) N;	


	dprintf("packet.cc: Packet::calcstate(): Normalized offset:  %d\n", chipoffset); 		
	int * ncodeword = codeset->RightShiftCodeword(codeword, chipoffset);
	dprintf("packet.cc: Packet::calcstate(): Normalized codeword:\n"); 		
	for (int count1 =0; count1 < A; count1++)
		for (int count = N -1; count >=0; count--)
		{
			::dprintf("%d ", ncodeword[count1*N + count]);	
			nstate[count1*N + count] += ncodeword[count1*N + count];
		}
	::dprintf("\n");	


	dprintf("packet.cc: Packet::calcstate(): Normalized state: "); 		
	for (int count1 =0; count1 < A; count1++)
		for (int count = N -1; count >=0; count--)
		{
			::dprintf("%d ", nstate[count1*N + count]);	
		}
	::dprintf("\n");	
	delete ncodeword;



}
