/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 Sept  2004
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <defs.h>
#include <packet.h>
#include <eventhandler.h>
#include <squeue.h>
#include <queue.h>
#include <scheduler.h>
#include <state.h>
#include <rng.h>
#include <trgen.h>
#include <trdump.h>
#include <txsched.h>
#include <intransit.h>
#include <codeset.h>
#include <debug.h>
#include <receiver.h>
#include <stateest.h>
#include <state.h>


extern Scheduler * sched;
extern InTransit * it;
extern TrDump * trdump;
extern TrGen * trgen;
extern char outputfile[20];
extern Codeset *codeset;
extern Receiver *rcvr;

extern int propSpeed;


StateEstimator::StateEstimator(int n, int a):EventHandler()
{
	A =a;
	N = n;

	totalreadings = 1000;
	double chiptime = (1.0/codeset->chiprate);	
	delay = N * chiptime; 

	totalerror = new double[totalreadings];
	for(int count =0; count < totalreadings; count++)
		totalerror[count] = 0;

	statechanges = 0;	
	oldstate = NULL;
	reading = 0;
	totalsamples =0;
	debug = 0;
}

StateEstimator::~StateEstimator()
{
}


void StateEstimator::process(Packet *p)
{


}


//The event handling function of the StateEstimator. It dequeues a packet and
//sends it to the next component.
void StateEstimator::handleevent(Packet *p)
{

	//Collect stats about the line state at regular intervals of time.
	dtprintf("stateest.cc: StateEstimator::handleevent(): Entered function\n");
	if(reading == totalreadings)
	{
		totalsamples++;
		reading =0;
		dtprintf("stateest.cc: StateEstimator::stimatestate(): Returning\n");
	}
	else
	{
		collectstats();
		reading++;
		sched->scheduleevent(this, NULL, delay);
	}
}



void StateEstimator::collectstats()
{

	int * newstate = estimatestate(estx, 0, 0, estx, 0, 0);
	dprintf("%.12f " , sched->currenttime);
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			dprintf("%d ", newstate[count1*N + count]);
		}
		dprintf("\n");
	}


	double errorsq = 0 ;
	double totalerrorsq = 0 ;
	for (int count1 =0; count1 < A; count1++)
	{
		for (int count =0; count < N; count++)
		{
			if(oldstate != NULL)
				errorsq = pow (newstate[count1*N + count] - oldstate[count1*N + count], 2);	
			totalerrorsq+=errorsq;	
		}
	}
	double error = pow(totalerrorsq, 0.5);
	dtprintf("Error: %.6f\n", error);
	if(error != lasterror)
		statechanges++;
	lasterror = error;
	totalerror[reading]+=error;


}

void StateEstimator::dumpstats()
{
		double avgerror = 0;
		dtprintf("Total samples: %d\n", totalsamples);
		for (int count =0; count < totalreadings; count++)
		{
			avgerror = totalerror[count]/totalsamples;
		}
		dtprintf("Average error: %.6f\n", avgerror);
		dtprintf("Average number of state changes: %.6f\n", (double)statechanges/totalsamples);
}


	
int  * StateEstimator::estimatestate(double ex, double ey, double ez, double tx, double ty, double tz)
{

	dtprintf("stateest.cc: StateEstimator::estimatestate(): Entered function\n");
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation point: %.6f, %.6f, %.6f\n", ex, ey, ez);
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Transmission point: %.6f, %.6f, %.6f\n", tx, ty, tz);
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation time: %.12f\n", sched->currenttime);


	double timeoffset = sched->currenttime;
	double chiptime = 1.0/codeset->chiprate;
	int chipoffset = int((timeoffset/chiptime) + 0.5);

	chipoffset = chipoffset % (int) N;	
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation time offset from normalized time (chips): %d\n", chipoffset);
		

	int * lstate = new int [A*N];
	for (int count = 0; count < A*N; count++)
		lstate[count]=0;
	Packet *p;
	while((p = it->traverse()) != NULL )
	{

		double head, tail;
		p->calcboundary(&head, &tail);
		if((head <= ex) && (tail > ex))
		{
			int * state = p->calcstate(ex, ey, ez, 0, 0);
			dtprintf("stateest.cc: StateEstimator::estimateetate: Packet: %d\n", p->id);
			dtprintf("stateest.cc: StateEstimator::estimatestate: Code:\n");
			for (int count = 0; count < A*N; count++)
				dprintf("%d ", state[count]);
			dprintf("\n");
			for (int count = 0; count < A*N; count++)
				lstate[count]+=state[count];		
			delete state;
		}

	}
	
	double chiplength = (1.0/codeset->chiprate) * propSpeed;	
	double offset = tx - ex;

	chipoffset =0;
	if (offset >= 0)
		chipoffset = int((offset/chiplength) + 0.5);
	else
		chipoffset = int((offset/chiplength) - 0.5);

	chipoffset = chipoffset % (int) N;	
	if(chipoffset < 0)
		chipoffset = -chipoffset;
		
	dtprintf("stateest.cc: StateEstimator::estimatestate(): Estimation distance offset (chips): %d\n", chipoffset);
	dprintf("\n");
	int * tstate = codeset->LeftShiftCodeword(lstate, chipoffset);	
	delete lstate;			

	return tstate;
}


